@@ -304,6 +304,87 @@ fn read_until() {
     assert_eq!(process.exit(true).unwrap(), true);
 }
 
+#[test]
+fn read_to_end() {
+    let mut cmd = Command::new("echo");
+    cmd.arg("Hello World");
+    let mut process = PtyProcess::spawn(cmd).unwrap();
+
+    let mut buf = Vec::new();
+    process.read_to_end(&mut buf).unwrap();
+    assert_eq!(&buf, b"Hello World\r\n");
+}
+
+#[cfg(not(target_os = "macos"))]
+#[test]
+fn read_to_end_after_delay() {
+    let mut cmd = Command::new("echo");
+    cmd.arg("Hello World");
+    let mut process = PtyProcess::spawn(cmd).unwrap();
+
+    thread::sleep(Duration::from_millis(500));
+
+    let mut buf = Vec::new();
+    process.read_to_end(&mut buf).unwrap();
+    assert_eq!(&buf, b"Hello World\r\n");
+}
+
+#[cfg(not(target_os = "macos"))]
+#[test]
+fn read_after_process_is_gone() {
+    let mut cmd = Command::new("echo");
+    cmd.arg("Hello World");
+    let mut process = PtyProcess::spawn(cmd).unwrap();
+
+    // after we check a status of child
+    // it should be marked DEAD.
+    assert_eq!(
+        process.wait().unwrap(),
+        WaitStatus::Exited(process.pid(), 0)
+    );
+
+    // Just in case; make a little delay
+    thread::sleep(Duration::from_millis(500));
+
+    let mut buf = vec![0; 128];
+    let n = process.read(&mut buf).unwrap();
+    assert_eq!(&buf[..n], b"Hello World\r\n");
+}
+
+#[cfg(not(target_os = "macos"))]
+#[test]
+fn read_to_end_after_process_is_gone() {
+    let mut cmd = Command::new("echo");
+    cmd.arg("Hello World");
+    let mut process = PtyProcess::spawn(cmd).unwrap();
+
+    // after we check a status of child
+    // it should be marked DEAD.
+    assert_eq!(
+        process.wait().unwrap(),
+        WaitStatus::Exited(process.pid(), 0)
+    );
+
+    // Just in case; make a little delay
+    thread::sleep(Duration::from_millis(500));
+
+    let mut buf = Vec::new();
+    process.read_to_end(&mut buf).unwrap();
+    assert_eq!(&buf, b"Hello World\r\n");
+}
+
+#[test]
+fn try_read_to_end() {
+    let mut cmd = Command::new("echo");
+    cmd.arg("Hello World");
+    let mut process = PtyProcess::spawn(cmd).unwrap();
+
+    let mut buf = vec![0; 128];
+    while process.try_read(&mut buf).unwrap() != Some(0) {}
+
+    assert_eq!(&buf[..13], b"Hello World\r\n");
+}
+
 #[test]
 fn continues_try_reads() {
     let mut cmd = Command::new("python3");
@@ -329,13 +410,33 @@ fn continues_try_reads() {
 }
 
 #[test]
+#[cfg(not(target_os = "macos"))]
 fn end_of_interact() {
     let mut p = PtyProcess::spawn(Command::new("ls")).unwrap();
     let status = p.interact().unwrap();
-    assert!(matches!(status, WaitStatus::Exited(_, 0)));
+
+    // It may be finished not only because process is done but
+    // also because it reached EOF.
+    assert!(matches!(
+        status,
+        WaitStatus::Exited(_, 0) | WaitStatus::StillAlive
+    ));
 
     // check that second spawn works
     let mut p = PtyProcess::spawn(Command::new("ls")).unwrap();
     let status = p.interact().unwrap();
-    assert!(matches!(status, WaitStatus::Exited(_, 0)));
+    assert!(matches!(
+        status,
+        WaitStatus::Exited(_, 0) | WaitStatus::StillAlive
+    ));
+}
+
+#[test]
+#[cfg(not(target_os = "macos"))]
+fn spawn_after_interact() {
+    let mut p = PtyProcess::spawn(Command::new("ls")).unwrap();
+    let _ = p.interact().unwrap();
+
+    let p = PtyProcess::spawn(Command::new("ls")).unwrap();
+    assert!(matches!(p.wait().unwrap(), WaitStatus::Exited(_, 0)));
 }
